"""
Module for useful functions that aren't available elsewhere and don't fit with Pymagicc.

When we say "available elsewhere and don't fit with Pymagicc" we mean that they are
not available to be included as a dependency of Pymagicc nor are they really/solely
related to running MAGICC. A good test of what belongs here is anything for which one
thinks, "I need this, I can't find it anywhere else and I would probably use it again
in another project". One day we may move all these utility functions to another
project to make it easier for ourselves and others to re-use them.
"""
from copy import deepcopy
import re
import datetime
import warnings


# Thank you https://stackoverflow.com/a/15448887/10473080
def _compile_replacement_regexp(rep_dict, case_insensitive=False):

    if case_insensitive:
        flags = re.DOTALL | re.IGNORECASE
    else:
        flags = re.DOTALL

    return re.compile(
        "|".join([re.escape(k) for k in sorted(rep_dict, key=len, reverse=True)]),
        flags=flags,
    )


def _multiple_replace(in_str, rep_dict, compiled_regexp):
    # To handle cases where compiled regexp is case insensitive, we can alter our
    # rep_dict keys to be all upper case cases and then replace by converting our
    # found groups to all upper case too. As long as our rep_dict values do not
    # change, the behaviour is as desired.
    rep_dict = {k.upper(): v for k, v in rep_dict.items()}
    out_str = compiled_regexp.sub(lambda x: rep_dict[x.group(0).upper()], in_str)
    return out_str


def _check_unused_substitutions(
    substitutions, inputs, unused_substitutions, case_insensitive
):
    _check_inputs = [inputs] if isinstance(inputs, str) else inputs
    _check_subs = list(substitutions.keys())
    if case_insensitive:
        _check_inputs = [v.upper() for v in _check_inputs]
        _check_subs = [v.upper() for v in _check_subs]

    unused = set(_check_inputs) - set(_check_subs)
    if not unused:
        return

    msg = "No substitution available for {}".format(unused)
    if unused_substitutions == "warn":
        warnings.warn(msg)
    elif unused_substitutions == "raise":
        raise ValueError(msg)
    else:
        raise ValueError("Invalid value for unused_substitutions, please see the docs")


def _check_duplicate_substitutions(substitutions):
    seen = set()
    duplicated = {}
    for key, value in substitutions.items():
        key_case_insensitive = key.upper()
        if key_case_insensitive in seen:
            duplicated.update(
                {
                    k: v
                    for k, v in substitutions.items()
                    if k.upper() == key_case_insensitive
                }
            )
        seen.add(key_case_insensitive)

    if duplicated:
        error_msg = "Duplicate case insensitive substitutions: {}".format(duplicated)
        raise ValueError(error_msg)


def apply_string_substitutions(
    inputs,
    substitutions,
    inverse=False,
    case_insensitive=False,
    unused_substitutions="ignore",
):
    """Apply a number of substitutions to a string(s).

    The substitutions are applied effectively all at once. This means that conflicting
    substitutions don't interact. Where substitutions are conflicting, the one which
    is longer takes precedance. This is confusing so we recommend that you look at
    the examples.

    Parameters
    ----------
    inputs : str, list of str
        The string(s) to which we want to apply the substitutions.

    substitutions : dict
        The substitutions we wish to make. The keys are the strings we wish to
        substitute, the values are the strings which we want to appear in the output
        strings.

    inverse : bool
        If True, do the substitutions the other way around i.e. use the keys as the
        strings we want to appear in the output strings and the values as the strings
        we wish to substitute.

    case_insensitive : bool
        If True, the substitutions will be made in a case insensitive way.

    unused_substitutions : {"ignore", "warn", "raise"}, default ignore
        Behaviour when one or more of the inputs does not have a corresponding
        substitution. If "ignore", nothing happens. If "warn", a warning is issued. If
        "raise", an error is raised. See the examples.

    Returns
    -------
    ``type(input)``
        The input with substitutions performed.

    Examples
    --------
    >>> apply_string_substitutions("Hello JimBob", {"Jim": "Bob"})
    'Hello BobBob'

    >>> apply_string_substitutions("Hello JimBob", {"Jim": "Bob"}, inverse=True)
    'Hello JimJim'

    >>> apply_string_substitutions(["Hello JimBob", "Jim says, 'Hi Bob'"], {"Jim": "Bob"})
    ['Hello BobBob', "Bob says, 'Hi Bob'"]

    >>> apply_string_substitutions(["Hello JimBob", "Jim says, 'Hi Bob'"], {"Jim": "Bob"}, inverse=True)
    ['Hello JimJim', "Jim says, 'Hi Jim'"]

    >>> apply_string_substitutions("Muttons Butter", {"M": "B", "Button": "Zip"})
    'Buttons Butter'
    # Substitutions don't cascade. If they did, Muttons would become Buttons, then the
    # substitutions "Button" --> "Zip" would be applied and we would end up with
    # "Zips Butter".

    >>> apply_string_substitutions("Muttons Butter", {"Mutton": "Gutter", "tt": "zz"})
    'Gutters Buzzer'
    # Longer substitutions take precedent. Hence Mutton becomes Gutter, not Muzzon.

    >>> apply_string_substitutions("Butter", {"buTTer": "Gutter"}, case_insensitive=True)
    'Gutter'

    >>> apply_string_substitutions("Butter", {"teeth": "tooth"})
    'Butter'

    >>> apply_string_substitutions("Butter", {"teeth": "tooth"}, unused_substitutions="ignore")
    'Butter'

    >>> apply_string_substitutions("Butter", {"teeth": "tooth"}, unused_substitutions="warn")
    ...pymagicc/utils.py:50: UserWarning: No substitution available for {'Butter'} warnings.warn(msg)
    'Butter'

    >>> apply_string_substitutions("Butter", {"teeth": "tooth"}, unused_substitutions="raise")
    ValueError: No substitution available for {'Butter'}
    """
    if inverse:
        substitutions = {v: k for k, v in substitutions.items()}

    # only possible to have conflicting substitutions when case insensitive
    if case_insensitive:
        _check_duplicate_substitutions(substitutions)

    if unused_substitutions != "ignore":
        _check_unused_substitutions(
            substitutions, inputs, unused_substitutions, case_insensitive
        )

    compiled_regexp = _compile_replacement_regexp(
        substitutions, case_insensitive=case_insensitive
    )

    inputs_return = deepcopy(inputs)
    if isinstance(inputs_return, str):
        inputs_return = _multiple_replace(inputs_return, substitutions, compiled_regexp)
    else:
        inputs_return = [
            _multiple_replace(v, substitutions, compiled_regexp) for v in inputs_return
        ]

    return inputs_return


def get_date_time_string():
    """
    Return a timestamp with current date and time.
    """
    now = datetime.datetime.now()
    return now.strftime("%Y-%m-%d %H:%M")



"""
This module contains all of the relevant definitions for handling MAGICC data.

When we store the data in csv's, we use `Data Packages <https://frictionlessdata.io/
docs/creating-tabular-data-packages-in-python/>`_. These store the data in an easy to
read csv file whilst providing comprehensive metadata describing the data (column
meanings and expected types) in the accompanying ``datapackage.json`` file. Please see
this metadata for further details.

For more details about how these constants are used, see the documentation of
``pymagicc.io``. In particular, the documentation of
``pymagicc.io.get_special_scen_code``, ``pymagicc.io.get_dattype_regionmode`` and
``pymagicc.io.get_region_order`` in :ref:`pymagicc.io`.
"""
from pathlib import Path
import warnings
import functools

import pandas as pd
from pandas_datapackage_reader import read_datapackage


from pymagicc.utils import apply_string_substitutions

DATA_HIERARCHY_SEPARATOR = "|"
"""str: String used to define different levels in our data hierarchies.

For example, "Emissions|CO2|Energy|Coal".

We copy this straight from pyam_ to maintain easy compatibility.
"""

path = Path(__file__).parent


_dtrm = read_datapackage(path, "magicc_dattype_regionmode_regions")

_region_cols = _dtrm.columns.to_series().apply(lambda x: x.startswith("region"))

DATTYPE_REGIONMODE_REGIONS = _dtrm.loc[:, ~_region_cols].copy()
""":obj:`pandas.DataFrame` Mapping between regions and whether a file is SCEN7 or not and the expected values of THISFILE_DATTYPE and THISFILE_REGIONMODE flags in MAGICC.
"""

DATTYPE_REGIONMODE_REGIONS["regions"] = [
    [r for r in raw if not pd.isnull(r)]
    for raw in _dtrm.loc[:, _region_cols].values.tolist()
]

MAGICC7_EMISSIONS_UNITS = read_datapackage(path, "magicc_emisssions_units")
""":obj:`pandas.DataFrame` Definitions of emissions variables and their expected units in MAGICC7.
"""

PART_OF_SCENFILE_WITH_EMISSIONS_CODE_0 = MAGICC7_EMISSIONS_UNITS[
    MAGICC7_EMISSIONS_UNITS["part_of_scenfile_with_emissions_code_0"]
]["magicc_variable"].tolist()
"""list: The emissions which are included in a SCEN file if the SCEN emms code is 0.

See documentation of ``pymagicc.io.get_special_scen_code`` for more details.
"""

PART_OF_SCENFILE_WITH_EMISSIONS_CODE_1 = MAGICC7_EMISSIONS_UNITS[
    MAGICC7_EMISSIONS_UNITS["part_of_scenfile_with_emissions_code_1"]
]["magicc_variable"].tolist()
"""list: The emissions which are included in a SCEN file if the SCEN emms code is 1.

See documentation of ``pymagicc.io.get_special_scen_code`` for more details.
"""

PART_OF_PRNFILE = MAGICC7_EMISSIONS_UNITS[MAGICC7_EMISSIONS_UNITS["part_of_prnfile"]][
    "magicc_variable"
].tolist()
"""list: The emissions which are included in a ``.prn`` file.
"""

MAGICC7_CONCENTRATIONS_UNITS = read_datapackage(path, "magicc_concentrations_units")
""":obj:`pandas.DataFrame` Definitions of concentrations variables and their expected units in MAGICC7.
"""


def get_magicc_region_to_openscm_region_mapping(inverse=False):
    """Get the mappings from MAGICC to OpenSCM regions.

    This is not a pure inverse of the other way around. For example, we never provide
    "GLOBAL" as a MAGICC return value because it's unnecesarily confusing when we also
    have "World". Fortunately MAGICC doesn't ever read the name "GLOBAL" so this
    shouldn't matter.

    Parameters
    ----------
    inverse : bool
        If True, return the inverse mappings i.e. MAGICC to OpenSCM mappings
    Returns
    -------
    dict
        Dictionary of mappings
    """
    world = "World"

    def get_openscm_replacement(in_region):
        if in_region in ("WORLD", "GLOBAL"):
            return world
        if in_region in ("BUNKERS",):
            return DATA_HIERARCHY_SEPARATOR.join([world, "Bunkers"])
        if in_region in ("OCEAN",):
            return DATA_HIERARCHY_SEPARATOR.join([world, "Ocean"])
        if in_region in ("LAND",):
            return DATA_HIERARCHY_SEPARATOR.join([world, "Land"])
        if in_region in ("N34",):
            return DATA_HIERARCHY_SEPARATOR.join([world, "El Nino N3.4"])
        if in_region in ("AMV",):
            return DATA_HIERARCHY_SEPARATOR.join([world, "North Atlantic Ocean"])
        elif in_region.startswith(("NH", "SH")):
            in_region = in_region.replace("-", "")
            hem = "Northern Hemisphere" if "NH" in in_region else "Southern Hemisphere"
            if in_region in ("NH", "SH"):
                return DATA_HIERARCHY_SEPARATOR.join([world, hem])

            land_ocean = "Land" if "LAND" in in_region else "Ocean"
            return DATA_HIERARCHY_SEPARATOR.join([world, hem, land_ocean])
        else:
            return DATA_HIERARCHY_SEPARATOR.join([world, in_region])

    # we generate the mapping dynamically, the first name in the list
    # is the one which will be used for inverse mappings
    _magicc_regions = [
        "WORLD",
        "GLOBAL",
        "OECD90",
        "ALM",
        "REF",
        "ASIA",
        "R5ASIA",
        "R5OECD",
        "R5REF",
        "R5MAF",
        "R5LAM",
        "R5.2ASIA",
        "R5.2OECD",
        "R5.2REF",
        "R5.2MAF",
        "R5.2LAM",
        "NHOCEAN",
        "SHOCEAN",
        "NHLAND",
        "SHLAND",
        "NH-OCEAN",
        "SH-OCEAN",
        "NH-LAND",
        "SH-LAND",
        "SH",
        "NH",
        "BUNKERS",
        "N34",
        "AMV",
        "OCEAN",
        "LAND",
    ]

    replacements = {}
    for magicc_region in _magicc_regions:
        openscm_region = get_openscm_replacement(magicc_region)
        # i.e. if we've already got a value for the inverse, we don't want to overwrite
        if (openscm_region in replacements.values()) and inverse:
            continue
        replacements[magicc_region] = openscm_region

    if inverse:
        return {v: k for k, v in replacements.items()}
    else:
        # R6 doesn't really exist, they're supposed to be the SSP database's R5.2
        # regions. The R6 regions appeared in a few SSP scenario files (but will
        # hopefully never be released) to this is just in case.
        for r6 in ["R6OECD90", "R6REF", "R6LAM", "R6MAF", "R6ASIA"]:
            replacements[r6] = "{}|{}".format(
                world, r6.replace("R6", "R5.2").replace("90", "")
            )
        return replacements


MAGICC_REGION_TO_OPENSCM_REGION_MAPPING = get_magicc_region_to_openscm_region_mapping()
"""dict: Mappings from MAGICC regions to OpenSCM regions"""

OPENSCM_REGION_TO_MAGICC_REGION_MAPPING = get_magicc_region_to_openscm_region_mapping(
    inverse=True
)
"""dict: Mappings from OpenSCM regions to MAGICC regions
"""


@functools.lru_cache(None)
def _apply_convert_magicc_to_openscm_regions(regions, inverse):
    if inverse:
        return apply_string_substitutions(
            regions,
            OPENSCM_REGION_TO_MAGICC_REGION_MAPPING,
            unused_substitutions="ignore",  # TODO: make this warn and see what happens
        )
    else:
        return apply_string_substitutions(
            regions,
            MAGICC_REGION_TO_OPENSCM_REGION_MAPPING,
            unused_substitutions="ignore",  # TODO: make this warn and see what happens
            case_insensitive=True,  # MAGICC regions are case insensitive
        )


def convert_magicc_to_openscm_regions(regions, inverse=False):
    """
    Convert MAGICC regions to OpenSCM regions

    Parameters
    ----------
    regions : list_like, str
        Regions to convert

    inverse : bool
        If True, convert the other way i.e. convert OpenSCM regions to MAGICC7
        regions

    Returns
    -------
    ``type(regions)``
        Set of converted regions
    """
    if isinstance(regions, (list, pd.Index)):
        return [_apply_convert_magicc_to_openscm_regions(r, inverse) for r in regions]
    else:
        return _apply_convert_magicc_to_openscm_regions(regions, inverse)


def get_magicc7_to_openscm_variable_mapping(inverse=False):
    """Get the mappings from MAGICC7 to OpenSCM variables.

    Parameters
    ----------
    inverse : bool
        If True, return the inverse mappings i.e. OpenSCM to MAGICC7 mappings

    Returns
    -------
    dict
        Dictionary of mappings
    """

    def get_openscm_replacement(in_var):
        if in_var.endswith("_INVERSE_EMIS"):
            prefix = "Inverse Emissions"
        elif in_var.endswith("_EMIS"):
            prefix = "Emissions"
        elif in_var.endswith("_CONC"):
            prefix = "Atmospheric Concentrations"
        elif in_var.endswith("_EFFRF"):
            prefix = "Effective Radiative Forcing"
        elif in_var.endswith("_RF"):
            prefix = "Radiative Forcing"
        elif in_var.endswith("_OT"):
            prefix = "Optical Thickness"
        else:
            raise ValueError("This shouldn't happen")

        variable = in_var.split("_")[0]
        # I hate edge cases
        if variable.endswith("EQ"):
            variable = variable.replace("EQ", " Equivalent")

        if "GHG" in variable:
            variable = variable.replace("GHG", "Greenhouse Gases")

        if "BIOMASSAER" in variable:
            variable = variable.replace(
                "BIOMASSAER", "Aerosols|Direct Effect|MAGICC AFOLU"
            )

        if "CO2CH4N2O" in variable:
            variable = variable.replace("CO2CH4N2O", "CO2, CH4 and N2O")

        aggregate_indicators = {
            "KYOTO": "Kyoto Gases",
            "FGASSUM": "F Gases",
            "MHALOSUM": "Montreal Protocol Halogen Gases",
        }
        for agg_indicator, long_name in aggregate_indicators.items():
            if variable.startswith(agg_indicator):
                stripped_var = variable.replace(agg_indicator, "")
                if stripped_var:
                    variable = DATA_HIERARCHY_SEPARATOR.join([stripped_var, long_name])
                else:
                    variable = long_name

        edge_case_B = variable.upper() in ("HCFC141B", "HCFC142B")
        if variable.endswith("I"):
            variable = DATA_HIERARCHY_SEPARATOR.join(
                [variable[:-1], "MAGICC Fossil and Industrial"]
            )
        elif variable.endswith("B") and not edge_case_B:
            variable = DATA_HIERARCHY_SEPARATOR.join([variable[:-1], "MAGICC AFOLU"])
        elif variable.endswith("T"):
            variable = variable[:-1]

        case_adjustments = {
            "SOX": "SOx",
            "NOX": "NOx",
            "CC4F8": "cC4F8",
            "HFC134A": "HFC134a",
            "HFC143A": "HFC143a",
            "HFC152A": "HFC152a",
            "HFC227EA": "HFC227ea",
            "HFC236FA": "HFC236fa",
            "HFC245FA": "HFC245fa",
            "HFC365MFC": "HFC365mfc",
            "HCFC141B": "HCFC141b",
            "HCFC142B": "HCFC142b",
            "CH3CCL3": "CH3CCl3",
            "CCL4": "CCl4",
            "CH3CL": "CH3Cl",
            "CH2CL2": "CH2Cl2",
            "CHCL3": "CHCl3",
            "CH3BR": "CH3Br",
            "HALON1211": "Halon1211",
            "HALON1301": "Halon1301",
            "HALON2402": "Halon2402",
            "HALON1202": "Halon1202",
            "SOLAR": "Solar",
            "VOLCANIC": "Volcanic",
            "EXTRA": "Extra",
        }
        variable = apply_string_substitutions(variable, case_adjustments)

        return DATA_HIERARCHY_SEPARATOR.join([prefix, variable])

    magicc7_suffixes = ["_EMIS", "_CONC", "_EFFRF", "_RF", "_OT", "_INVERSE_EMIS"]
    magicc7_base_vars = MAGICC7_EMISSIONS_UNITS.magicc_variable.tolist() + [
        "SOLAR",
        "VOLCANIC",
        "CO2EQ",
        "KYOTOCO2EQ",
        "FGASSUMHFC134AEQ",
        "MHALOSUMCFC12EQ",
        "GHG",
        "KYOTOGHG",
        "FGASSUM",
        "MHALOSUM",
        "BIOMASSAER",
        "CO2CH4N2O",
        "EXTRA",
    ]
    magicc7_vars = [
        base_var + suffix
        for base_var in magicc7_base_vars
        for suffix in magicc7_suffixes
    ]

    replacements = {m7v: get_openscm_replacement(m7v) for m7v in magicc7_vars}

    rf_updates = {
        "TOTAL_INCLVOLCANIC_RF": "Radiative Forcing",
        "VOLCANIC_ANNUAL_RF": "Radiative Forcing|Volcanic",
        "TOTAL_ANTHRO_RF": "Radiative Forcing|Anthropogenic",
        "TOTAER_DIR_RF": "Radiative Forcing|Aerosols|Direct Effect",
        "CLOUD_TOT_RF": "Radiative Forcing|Aerosols|Indirect Effect",
        "OCI_RF": "Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC Fossil and Industrial",
        "BCI_RF": "Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC Fossil and Industrial",
        "SOXI_RF": "Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC Fossil and Industrial",
        "NOXI_RF": "Radiative Forcing|Aerosols|Direct Effect|NOx|MAGICC Fossil and Industrial",
        "NH3I_RF": "Radiative Forcing|Aerosols|Direct Effect|NH3|MAGICC Fossil and Industrial",
        "MINERALDUST_RF": "Radiative Forcing|Aerosols|Direct Effect|Mineral Dust",
        "STRATOZ_RF": "Radiative Forcing|Stratospheric Ozone",
        "TROPOZ_RF": "Radiative Forcing|Tropospheric Ozone",
        "CH4OXSTRATH2O_RF": "Radiative Forcing|CH4 Oxidation Stratospheric H2O",
        "LANDUSE_RF": "Radiative Forcing|Land-use Change",
        "BCSNOW_RF": "Radiative Forcing|Black Carbon on Snow",
        "AIR_CIRRUS_RF": "Radiative Forcing|Aviation|Cirrus",
        "AIR_CONTRAIL_RF": "Radiative Forcing|Aviation|Contrail",
        "AIR_CONTRAILANDCIRRUS_RF": "Radiative Forcing|Aviation|Contrail and Cirrus",
        "AIR_H2O_RF": "Radiative Forcing|Aviation|H2O",
    }
    rf_updates_with_effrf = {
        **rf_updates,
        **{
            k.replace("_RF", "_EFFRF"): v.replace(
                "Radiative Forcing", "Effective Radiative Forcing"
            )
            for k, v in rf_updates.items()
        },
    }
    replacements.update(rf_updates_with_effrf)
    replacements.update(
        {
            "SURFACE_TEMP": "Surface Temperature",
            "CO2PF_EMIS": "Land to Air Flux|CO2|MAGICC Permafrost",
            "CH4PF_EMIS": "Land to Air Flux|CH4|MAGICC Permafrost",
            "SLR_TOT": "Sea Level Rise",
        }
    )

    agg_ocean_heat_top = "Aggregated Ocean Heat Content"
    heat_content_aggreg_depths = {
        "HEATCONTENT_AGGREG_DEPTH{}".format(i): "{}{}Depth {}".format(
            agg_ocean_heat_top, DATA_HIERARCHY_SEPARATOR, i
        )
        for i in range(1, 4)
    }
    replacements.update(heat_content_aggreg_depths)
    replacements.update({"HEATCONTENT_AGGREG_TOTAL": agg_ocean_heat_top})

    ocean_temp_layer = {
        "OCEAN_TEMP_LAYER_{0:03d}".format(i): "Ocean Temperature{}Layer {}".format(
            DATA_HIERARCHY_SEPARATOR, i
        )
        for i in range(1, 999)
    }
    replacements.update(ocean_temp_layer)

    if inverse:
        return {v: k for k, v in replacements.items()}
    else:
        # these come from MAGICC's output
        one_way_replacements = {
            "CO2T_EMIS": "Emissions|CO2",
            "CH4T_EMIS": "Emissions|CH4",
            "N2OT_EMIS": "Emissions|N2O",
            "SURFACE_TEMP_SUBANNUAL": "Surface Temperature",
        }
        replacements.update(one_way_replacements)
        return replacements


MAGICC7_TO_OPENSCM_VARIABLES_MAPPING = get_magicc7_to_openscm_variable_mapping()
"""dict: Mappings from MAGICC7 variables to OpenSCM variables
"""

OPENSCM_TO_MAGICC7_VARIABLES_MAPPING = get_magicc7_to_openscm_variable_mapping(
    inverse=True
)
"""dict: Mappings from OpenSCM variables to MAGICC7 variables
"""


@functools.lru_cache(None)
def _apply_convert_magicc7_to_openscm_variables(v, inverse):
    if inverse:
        return apply_string_substitutions(
            v,
            OPENSCM_TO_MAGICC7_VARIABLES_MAPPING,
            unused_substitutions="ignore",  # TODO: make this warn and see what happens
        )
    else:
        return apply_string_substitutions(
            v,
            MAGICC7_TO_OPENSCM_VARIABLES_MAPPING,
            unused_substitutions="ignore",  # TODO: make this warn and see what happens
            case_insensitive=True,  # MAGICC variables are case insensitive
        )


def convert_magicc7_to_openscm_variables(variables, inverse=False):
    """
    Convert MAGICC7 variables to OpenSCM variables

    Parameters
    ----------
    variables : list_like, str
        Variables to convert

    inverse : bool
        If True, convert the other way i.e. convert OpenSCM variables to MAGICC7
        variables

    Returns
    -------
    ``type(variables)``
        Set of converted variables
    """
    if isinstance(variables, (list, pd.Index)):
        return [
            _apply_convert_magicc7_to_openscm_variables(v, inverse) for v in variables
        ]
    else:
        return _apply_convert_magicc7_to_openscm_variables(variables, inverse)


def get_magicc6_to_magicc7_variable_mapping(inverse=False):
    """Get the mappings from MAGICC6 to MAGICC7 variables.

    Note that this mapping is not one to one. For example, "HFC4310", "HFC43-10" and
    "HFC-43-10" in MAGICC6 both map to "HFC4310" in MAGICC7 but "HFC4310" in
    MAGICC7 maps back to "HFC4310".

    Note that HFC-245fa was mistakenly labelled as HFC-245ca in MAGICC6. In reality,
    they are not the same thing. However, the MAGICC6 labelling was merely a typo so
    the mapping between the two is one-to-one.

    Parameters
    ----------
    inverse : bool
        If True, return the inverse mappings i.e. MAGICC7 to MAGICC6 mappings

    Returns
    -------
    dict
        Dictionary of mappings
    """
    # we generate the mapping dynamically, the first name in the list
    # is the one which will be used for inverse mappings
    magicc6_simple_mapping_vars = [
        "KYOTO-CO2EQ",
        "CO2I",
        "CO2B",
        "CH4",
        "N2O",
        "BC",
        "OC",
        "SOx",
        "NOx",
        "NMVOC",
        "CO",
        "SF6",
        "NH3",
        "CF4",
        "C2F6",
        "HFC4310",
        "HFC43-10",
        "HFC-43-10",
        "HFC4310",
        "HFC134a",
        "HFC143a",
        "HFC227ea",
        "CCl4",
        "CH3CCl3",
        "HFC245fa",
        "Halon 1211",
        "Halon 1202",
        "Halon 1301",
        "Halon 2402",
        "Halon1211",
        "Halon1202",
        "Halon1301",
        "Halon2402",
        "CH3Br",
        "CH3Cl",
        "C6F14",
    ]

    magicc6_sometimes_hyphen_vars = [
        "CFC-11",
        "CFC-12",
        "CFC-113",
        "CFC-114",
        "CFC-115",
        "HCFC-22",
        "HFC-23",
        "HFC-32",
        "HFC-125",
        "HFC-134a",
        "HFC-143a",
        "HCFC-141b",
        "HCFC-142b",
        "HFC-227ea",
        "HFC-245fa",
    ]
    magicc6_sometimes_hyphen_vars = [
        v.replace("-", "") for v in magicc6_sometimes_hyphen_vars
    ] + magicc6_sometimes_hyphen_vars

    magicc6_sometimes_underscore_vars = [
        "HFC43_10",
        "CFC_11",
        "CFC_12",
        "CFC_113",
        "CFC_114",
        "CFC_115",
        "HCFC_22",
        "HCFC_141b",
        "HCFC_142b",
    ]
    magicc6_sometimes_underscore_replacements = {
        v: v.replace("_", "") for v in magicc6_sometimes_underscore_vars
    }

    special_case_replacements = {
        "FossilCO2": "CO2I",
        "OtherCO2": "CO2B",
        "MCF": "CH3CCL3",
        "CARB_TET": "CCL4",
        "MHALOSUMCFC12EQ": "MHALOSUMCFC12EQ",  # special case to avoid confusion with MCF
    }

    one_way_replacements = {"HFC-245ca": "HFC245FA", "HFC245ca": "HFC245FA"}

    all_possible_magicc6_vars = (
        magicc6_simple_mapping_vars
        + magicc6_sometimes_hyphen_vars
        + magicc6_sometimes_underscore_vars
        + list(special_case_replacements.keys())
        + list(one_way_replacements.keys())
    )
    replacements = {}
    for m6v in all_possible_magicc6_vars:
        if m6v in special_case_replacements:
            replacements[m6v] = special_case_replacements[m6v]
        elif (
            m6v in magicc6_sometimes_underscore_vars and not inverse
        ):  # underscores one way
            replacements[m6v] = magicc6_sometimes_underscore_replacements[m6v]
        elif (m6v in one_way_replacements) and not inverse:
            replacements[m6v] = one_way_replacements[m6v]
        else:
            m7v = m6v.replace("-", "").replace(" ", "").upper()
            # i.e. if we've already got a value for the inverse, we don't
            # want to overwrite it
            if (m7v in replacements.values()) and inverse:
                continue
            replacements[m6v] = m7v

    if inverse:
        return {v: k for k, v in replacements.items()}
    else:
        return replacements


MAGICC6_TO_MAGICC7_VARIABLES_MAPPING = get_magicc6_to_magicc7_variable_mapping()
"""dict: Mappings from MAGICC6 variables to MAGICC7 variables
"""

MAGICC7_TO_MAGICC6_VARIABLES_MAPPING = get_magicc6_to_magicc7_variable_mapping(
    inverse=True
)
"""dict: Mappings from MAGICC7 variables to MAGICC6 variables
"""


@functools.lru_cache(None)
def _apply_convert_magicc6_to_magicc7_variables(variables, inverse):
    def hfc245ca_included(variables):
        variables = [variables] if isinstance(variables, str) else variables
        return any([v.replace("-", "").lower() == "hfc245ca" for v in variables])

    if hfc245ca_included(variables):
        error_msg = (
            "HFC245ca wasn't meant to be included in MAGICC6. Renaming to HFC245fa."
        )
        warnings.warn(error_msg)

    if inverse:
        return apply_string_substitutions(
            variables,
            MAGICC7_TO_MAGICC6_VARIABLES_MAPPING,
            unused_substitutions="ignore",  # TODO: make this warn and see what happens
            case_insensitive=True,  # MAGICC variables are case insensitive
        )
    else:
        return apply_string_substitutions(
            variables,
            MAGICC6_TO_MAGICC7_VARIABLES_MAPPING,
            unused_substitutions="ignore",  # TODO: make this warn and see what happens
            case_insensitive=True,  # MAGICC variables are case insensitive
        )


def convert_magicc6_to_magicc7_variables(variables, inverse=False):
    """
    Convert MAGICC6 variables to MAGICC7 variables

    Parameters
    ----------
    variables : list_like, str
        Variables to convert

    inverse : bool
        If True, convert the other way i.e. convert MAGICC7 variables to MAGICC6
        variables

    Raises
    ------
    ValueError
        If you try to convert HFC245ca, or some variant thereof, you will get a
        ValueError. The reason is that this variable was never meant to be included in
        MAGICC6, it was just an accident. See, for example, the text in the
        description section of ``pymagicc/MAGICC6/run/HISTRCP_HFC245fa_CONC.IN``:
        "...HFC245fa, rather than HFC245ca, is the actually used isomer.".

    Returns
    -------
    ``type(variables)``
        Set of converted variables
    """
    if isinstance(variables, (list, pd.Index)):
        return [
            _apply_convert_magicc6_to_magicc7_variables(v, inverse) for v in variables
        ]
    else:
        return _apply_convert_magicc6_to_magicc7_variables(variables, inverse)


def get_pint_to_fortran_safe_units_mapping(inverse=False):
    """Get the mappings from Pint to Fortran safe units.

    Fortran can't handle special characters like "^" or "/" in names, but we need
    these in Pint. Conversely, Pint stores variables with spaces by default e.g. "Mt
    CO2 / yr" but we don't want these in the input files as Fortran is likely to think
    the whitespace is a delimiter.

    Parameters
    ----------
    inverse : bool
        If True, return the inverse mappings i.e. Fortran safe to Pint mappings

    Returns
    -------
    dict
        Dictionary of mappings
    """
    replacements = {"^": "super", "/": "per", " ": ""}
    if inverse:
        replacements = {v: k for k, v in replacements.items()}
        # mapping nothing to something is obviously not going to work in the inverse
        # hence remove
        replacements.pop("")

    return replacements


PINT_TO_FORTRAN_SAFE_UNITS_MAPPING = get_pint_to_fortran_safe_units_mapping()
"""dict: mappings required to make Pint units Fortran safe.
"""

FORTRAN_SAFE_TO_PINT_UNITS_MAPPING = get_pint_to_fortran_safe_units_mapping(
    inverse=True
)
"""dict: mappings required to convert our Fortran safe units to Pint.
"""


def convert_pint_to_fortran_safe_units(units, inverse=False):
    """
    Convert Pint units to Fortran safe units

    Parameters
    ----------
    units : list_like, str
        Units to convert

    inverse : bool
        If True, convert the other way i.e. convert Fortran safe units to Pint units

    Returns
    -------
    ``type(units)``
        Set of converted units
    """
    if inverse:
        return apply_string_substitutions(units, FORTRAN_SAFE_TO_PINT_UNITS_MAPPING)
    else:
        return apply_string_substitutions(units, PINT_TO_FORTRAN_SAFE_UNITS_MAPPING)



import shutil
import subprocess
import warnings
from os import listdir, makedirs
from os.path import basename, dirname, exists, join, isfile, abspath
from tempfile import mkdtemp
from dateutil.relativedelta import relativedelta
from copy import deepcopy


import numpy as np
import f90nml
import pandas as pd
from scmdata import df_append


from .config import config, _wine_installed
from .scenarios import zero_emissions
from .utils import get_date_time_string
from .io import (
    MAGICCData,
    NoReaderWriterError,
    InvalidTemporalResError,
    read_cfg_file,
    _get_openscm_var_from_filepath,
)


IS_WINDOWS = config["is_windows"]


class WineNotInstalledError(Exception):
    """Exception raised if wine is not installed but is required"""


def _copy_files(source, target):
    """
    Copy all the files in source directory to target.

    Ignores subdirectories.
    """
    source_files = listdir(source)
    if not exists(target):
        makedirs(target)
    for filename in source_files:
        full_filename = join(source, filename)
        if isfile(full_filename):
            shutil.copy(full_filename, target)


def _clean_value(v):
    if isinstance(v, str):
        return v.strip()
    elif isinstance(v, list):
        if isinstance(v[0], str):
            return [i.replace("\0", "").strip().replace("\n", "") for i in v]
    return v


class MAGICCBase(object):
    """
    Provides access to the MAGICC binary and configuration.

    To enable multiple MAGICC 'setups' to be configured independently,
    the MAGICC directory containing the input files, configuration
    and binary is copied to a new folder. The configuration in this
    MAGICC copy can then be edited without impacting other instances or your
    original MAGICC distribution.

    A ``MAGICC`` instance first has to be setup by calling
    ``create_copy``. If many model runs are being performed this step only has
    to be performed once. The ``run`` method can then be called many times
    without re-copying the files each time. Between each call to ``run``, the
    configuration files can be updated to perform runs with different
    configurations.

    Parameters
    ----------
    root_dir : str
        If ``root_dir`` is supplied, an existing MAGICC 'setup' is
        used.
    """

    version = None
    _scen_file_name = "SCENARIO.SCEN7"

    def __init__(self, root_dir=None, strict=True):
        """
        Initialise

        Parameters
        ----------
        root_dir : str
            Root directory of the MAGICC package. If ``None``, a temporary
            copy of MAGICC is made based on the result of `
            `self.get_exectuable()``.
        strict: bool
            If True, enforce the configuration checks, otherwise a warning
            is raised if any invalid configuration is found and the run is
            continued. Setting ``strict=False`` is only recommended for
            experienced users of MAGICC.
        """
        self.root_dir = root_dir
        self.config = None
        self.executable = self.get_executable()
        self.strict = strict

        if root_dir is not None:
            self.is_temp = False
        else:
            # Create a temp directory
            self.is_temp = True

    def __enter__(self):
        if self.is_temp and self.run_dir is None:
            self.create_copy()
        return self

    def __exit__(self, *args, **kwargs):
        self.remove_temp_copy()

    def create_copy(self):
        """
        Initialises a temporary directory structure and copy of MAGICC
        configuration files and binary.
        """
        if self.executable is None or not isfile(self.executable):
            raise FileNotFoundError(
                "Could not find MAGICC{} executable: {}".format(
                    self.version, self.executable
                )
            )
        if self.is_temp:
            assert (
                self.root_dir is None
            ), "A temp copy for this instance has already been created"
            self.root_dir = mkdtemp(prefix="pymagicc-")

        if exists(self.run_dir):
            raise Exception("A copy of MAGICC has already been created.")
        if not exists(self.root_dir):
            makedirs(self.root_dir)

        exec_dir = basename(self.original_dir)

        # Copy a subset of folders from the MAGICC `original_dir`
        # Also copy anything which is in the root of the MAGICC distribution
        # Assumes that the MAGICC binary is in a folder one level below the root
        # of the MAGICC distribution. i.e. /run/magicc.exe or /bin/magicc
        dirs_to_copy = [".", "bin", "run"]
        # Check that the executable is in a valid sub directory
        assert exec_dir in dirs_to_copy, "binary must be in bin/ or run/ directory"

        for d in dirs_to_copy:
            source_dir = abspath(join(self.original_dir, "..", d))
            if exists(source_dir):
                _copy_files(source_dir, join(self.root_dir, d))

        # Create an empty out dir
        # MAGICC assumes that the 'out' directory already exists
        makedirs(join(self.root_dir, "out"))

        # Create basic configuration files so magicc can run
        self.set_years()
        self.set_config()

    @property
    def binary_name(self):
        """
        Name of the MAGICC binary file

        Returns
        -------
        str
            Name of the binary file
        """
        return basename(self.executable)

    @property
    def original_dir(self):
        """
        Directory of the MAGICC package.

        This is the directory which contains the ``run`` and ``out`` folders.

        Returns
        -------
        str
            Path of the MAGICC package
        """
        return dirname(self.executable)

    @property
    def run_dir(self):
        """
        Run directory of the MAGICC package.

        This path always ends in ``run``.

        Returns
        -------
        str
            Path of the run directory
        """
        if self.root_dir is None:
            return None
        return join(self.root_dir, "run")

    @property
    def out_dir(self):
        """
        Output directory of the MAGICC package.

        This path always ends in ``out``.

        Returns
        -------
        str
            Path of the output directory
        """
        if self.root_dir is None:
            return None
        return join(self.root_dir, "out")

    @property
    def default_config(self):
        """
        Default configuration for a run

        Returns
        -------
        :obj:`f90nml.Namelist`
            Namelist object containing the default configuration
        """
        base = f90nml.read(join(self.run_dir, "MAGCFG_DEFAULTALL.CFG"))
        user = f90nml.read(join(self.run_dir, "MAGCFG_USER.CFG"))
        self._default_config = deepcopy(base)

        def _deep_update(b, o):
            for k, v in o.items():
                if isinstance(v, dict):
                    _deep_update(b[k], v)
                else:
                    b.update(o)

        _deep_update(self._default_config, user)

        return self._default_config

    def run(self, scenario=None, only=None, debug=False, **kwargs):
        """
        Run MAGICC and parse the output.

        As a reminder, putting ``out_parameters=1`` will cause MAGICC to write out its
        parameters into ``out/PARAMETERS.OUT`` and they will then be read into
        ``output.metadata["parameters"]`` where ``output`` is the returned object.

        Any logged output from running magicc will be in``output.metadata["stderr"]``.
        For MAGICC7 and above, The level of logging can be controlled with the
        ``debug`` argument.

        Parameters
        ----------
        scenario : :obj:`pymagicc.io.MAGICCData`
            Scenario to run. If None MAGICC will simply run with whatever config has
            already been set.

        only : list of str
            If not None, only extract variables in this list.

        debug: {True, False, "verbose"}
            If true, MAGICC will run in debug mode with the maximum amount of logging.
            If "verbose", MAGICC will be run in verbose mode.

        kwargs
            Other config values to pass to MAGICC for the run

        Returns
        -------
        :obj:`pymagicc.io.MAGICCData`
            MAGICCData object containing that data in its ``df`` attribute and
            metadata and parameters (depending on the value of ``include_parameters``)
            in its ``metadata`` attribute.

        Raises
        ------
        ValueError
            If no output is found which matches the list specified in ``only``.

        subprocess.CalledProcessError
            If MAGICC fails to run. Check the 'stderr' key of the result's `metadata`
            attribute to inspect the results output from MAGICC.

        ValueError
            The user attempts to use ``debug`` with MAGICC6
        """
        if not exists(self.root_dir):
            raise FileNotFoundError(self.root_dir)

        if self.executable is None:
            raise ValueError(
                "MAGICC executable not found, try setting an environment variable `MAGICC_EXECUTABLE_{}=/path/to/binary`".format(
                    self.version
                )
            )

        if scenario is not None:
            kwargs = self.set_emission_scenario_setup(scenario, kwargs)

        yr_config = {}
        if "startyear" in kwargs:
            yr_config["startyear"] = kwargs.pop("startyear")
        if "endyear" in kwargs:
            yr_config["endyear"] = kwargs.pop("endyear")
        if yr_config:
            self.set_years(**yr_config)

        # should be able to do some other nice metadata stuff re how magicc was run
        # etc. here
        kwargs.setdefault("rundate", get_date_time_string())

        self.update_config(**kwargs)

        self.check_config()

        exec_dir = basename(self.original_dir)
        command = [join(self.root_dir, exec_dir, self.binary_name)]
        if self.version >= 7:
            if debug == "verbose":
                command.append("--verbose")
            elif debug:
                command.append("--debug")
        elif debug:
            raise ValueError("MAGICC6 has no debug capability")

        if not IS_WINDOWS and self.binary_name.endswith(".exe"):  # pragma: no cover
            if not _wine_installed:
                raise WineNotInstalledError(
                    "Wine is not installed but is required to run `.exe` binaries"
                )
            command.insert(0, "wine")

        # On Windows shell=True is required.
        try:
            res = subprocess.run(
                command,
                check=True,
                capture_output=True,
                cwd=self.run_dir,
                shell=IS_WINDOWS,
            )
        except subprocess.CalledProcessError as exc:
            print("stderr:\n{}".format(exc.stderr.decode()))
            raise exc

        outfiles = self._get_output_filenames()

        read_cols = {"climate_model": ["MAGICC{}".format(self.version)]}
        if scenario is not None:
            read_cols["model"] = scenario["model"].unique().tolist()
            read_cols["scenario"] = scenario["scenario"].unique().tolist()
        else:
            read_cols.setdefault("model", ["unspecified"])
            read_cols.setdefault("scenario", ["unspecified"])

        mdata = []
        for filepath in outfiles:
            try:
                openscm_var = _get_openscm_var_from_filepath(filepath)
                if only is None or openscm_var in only:
                    tempdata = MAGICCData(
                        join(self.out_dir, filepath), columns=deepcopy(read_cols)
                    )
                    mdata.append(tempdata)

            except (NoReaderWriterError, InvalidTemporalResError):
                # TODO: something like warnings.warn("Could not read {}".format(filepath))
                continue

        if not mdata and only is not None:
            raise ValueError("No output found for only={}".format(only))

        if not mdata:
            if self.strict:
                raise ValueError("No output found. Check configuration")
            else:
                # No data was loaded return an empty MAGICCData object
                mdata = MAGICCData(
                    data={},
                    columns={
                        "model": [],
                        "unit": [],
                        "variable": [],
                        "region": [],
                        "scenario": [],
                    },
                )
        else:
            mdata = df_append(mdata)

        try:
            run_paras = self.read_parameters()
            self.config = run_paras
            mdata.metadata["parameters"] = run_paras
        except FileNotFoundError:
            pass

        mdata.metadata["stderr"] = res.stderr.decode("ascii")
        levels_to_warn = ["WARNING", "ERROR", "FATAL"]
        for l in levels_to_warn:
            if l in mdata.metadata["stderr"]:
                warnings.warn(
                    "magicc logged a {} message. Check the 'stderr' key of the "
                    "result's `metadata` attribute.".format(l)
                )

        return mdata

    def _get_output_filenames(self):
        outfiles = [f for f in listdir(self.out_dir) if f != "PARAMETERS.OUT"]

        bin_out = [
            f.split(".")[0]
            for f in outfiles
            if f.startswith("DAT_") and f.endswith(".BINOUT")
        ]

        extras = []
        for f in outfiles:
            var_name, ext = f.split(".")
            if ext != "BINOUT" and var_name not in bin_out:
                extras.append(f)

        return [f + ".BINOUT" for f in bin_out] + extras

    def _check_failed(self, msg):
        if self.strict:
            raise ValueError(msg)
        else:
            warnings.warn(msg)

    def check_config(self):
        """Check that our MAGICC ``.CFG`` files are set to safely work with PYMAGICC

        For further detail about why this is required, please see :ref:`MAGICC flags`.

        Raises
        ------
        ValueError
            If we are not certain that the config written by PYMAGICC will overwrite
            all other config i.e. that there will be no unexpected behaviour. A
            ValueError will also be raised if the user tries to use more than one
            scenario file.
        """
        cfg_error_msg = (
            "PYMAGICC is not the only tuning model that will be used by "
            "`MAGCFG_USER.CFG`: your run is likely to fail/do odd things"
        )
        emisscen_error_msg = (
            "You have more than one `FILE_EMISSCEN_X` flag set. Using more than "
            "one emissions scenario is hard to debug and unnecessary with "
            "Pymagicc's Dataframe scenario input. Please combine all your "
            "scenarios into one Dataframe with Pymagicc and Pandas, then feed "
            "this single Dataframe into Pymagicc's run API."
        )

        nml_to_check = "nml_allcfgs"
        usr_cfg = read_cfg_file(join(self.run_dir, "MAGCFG_USER.CFG"))
        for k in usr_cfg[nml_to_check]:
            if k.startswith("file_tuningmodel"):
                first_tuningmodel = k in ["file_tuningmodel", "file_tuningmodel_1"]
                if first_tuningmodel:
                    if usr_cfg[nml_to_check][k] != "PYMAGICC":
                        self._check_failed(cfg_error_msg)
                elif usr_cfg[nml_to_check][k] not in ["USER", ""]:
                    self._check_failed(cfg_error_msg)

            elif k.startswith("file_emisscen_"):
                if usr_cfg[nml_to_check][k] not in ["NONE", ""]:
                    self._check_failed(emisscen_error_msg)

        self._check_config()

    def write(self, mdata, name):
        """Write an input file to disk

        Parameters
        ----------
        mdata : :obj:`pymagicc.io.MAGICCData`
            A MAGICCData instance with the data to write

        name : str
            The name of the file to write. The file will be written to the MAGICC
            instance's run directory i.e. ``self.run_dir``
        """
        mdata.write(join(self.run_dir, name), self.version)

    def read_parameters(self):
        """
        Read a parameters.out file

        Returns
        -------
        dict
            A dictionary containing all the configuration used by MAGICC
        """
        param_fname = join(self.out_dir, "PARAMETERS.OUT")

        if not exists(param_fname):
            raise FileNotFoundError("No PARAMETERS.OUT found")

        with open(param_fname) as nml_file:
            parameters = dict(f90nml.read(nml_file))
            for group in ["nml_years", "nml_allcfgs", "nml_outputcfgs"]:
                parameters[group] = dict(parameters[group])
                for k, v in parameters[group].items():
                    parameters[group][k] = _clean_value(v)
                parameters[group.replace("nml_", "")] = parameters.pop(group)
            self.config = parameters
        return parameters

    def remove_temp_copy(self):
        """
        Removes a temporary copy of the MAGICC version shipped with Pymagicc.
        """
        if self.is_temp and self.root_dir is not None:
            shutil.rmtree(self.root_dir)
            self.root_dir = None

    def set_config(
        self,
        filename="MAGTUNE_PYMAGICC.CFG",
        top_level_key="nml_allcfgs",
        conflict="warn",
        **kwargs,
    ):
        """
        Create a configuration file for MAGICC.

        Writes a fortran namelist in run_dir.

        Parameters
        ----------
        filename : str
            Name of configuration file to write

        top_level_key : str
            Name of namelist to be written in the
            configuration file

        conflict : {'warn', 'ignore'}
            If 'warn', when a flag needs to be replaced by a different name (because,
            for example, the flag name changed between MAGICC versions), a warning is
            raised. If 'ignore', no warning is raised when a replacement is required.

        kwargs
            Other parameters to pass to the configuration file. No
            validation on the parameters is performed.

        Returns
        -------
        dict
            The contents of the namelist which was written to file

        Warning
        -------
            If a key is renamed, a warning is raised

        Raises
        ------
        ValueError
            An invalid value for ``conflict`` is supplied
        """
        kwargs = self._format_config(kwargs)

        fname = join(self.run_dir, filename)
        conf = {top_level_key: kwargs}
        conf = self._fix_legacy_keys(conf, conflict=conflict)
        f90nml.write(conf, fname, force=True)

        return conf

    def update_config(
        self,
        filename="MAGTUNE_PYMAGICC.CFG",
        top_level_key="nml_allcfgs",
        conflict="warn",
        **kwargs,
    ):
        """Updates a configuration file for MAGICC

        Updates the contents of a fortran namelist in the run directory,
        creating a new namelist if none exists.

        Parameters
        ----------
        filename : str
            Name of configuration file to write

        top_level_key : str
            Name of namelist to be written in the
            configuration file

        conflict : {'warn', 'ignore'}
            If 'warn', when a flag needs to be replaced by a different name (because,
            for example, the flag name changed between MAGICC versions), a warning is
            raised. If 'ignore', no warning is raised when a replacement is required.

        kwargs
            Other parameters to pass to the configuration file. No
            validation on the parameters is performed.

        Returns
        -------
        dict
            The contents of the namelist which was written to file

        Warning
        -------
            If a key is renamed, a warning is raised

        Raises
        ------
        ValueError
            An invalid value for ``conflict`` is supplied
        """
        kwargs = self._format_config(kwargs)
        fname = join(self.run_dir, filename)

        if exists(fname):
            conf = f90nml.read(fname)
        else:
            conf = {top_level_key: {}}

        conf[top_level_key].update(kwargs)
        conf = self._fix_legacy_keys(conf, conflict=conflict)
        f90nml.write(conf, fname, force=True)

        return conf

    def _fix_legacy_keys(self, conf, conflict="warn"):
        """
        Go through config and fix any keys which are misnamed.

        For example, fix any keys which have been renamed between MAGICC versions to
        match the new names.

        Parameters
        ----------
        conf :obj:`f90nml.Namelist`
            Configuration to check

        conflict : {'warn', 'ignore'}
            If 'warn', when a conflict is found, a warning is raised. If 'ignore', no
            warning is raised when a conflict is found.

        Returns
        -------
        :obj:`f90nml.Namelist`
            Configuration with updated keys

        Warning
        -------
            If a key is renamed, a warning is raised

        Raises
        ------
        ValueError
            An invalid value for ``conflict`` is supplied
        """
        valid_conflicts = ["warn", "ignore"]
        if conflict not in valid_conflicts:
            raise ValueError("`conflict` must be one of: {}".format(valid_conflicts))

        cfg_key = "nml_allcfgs"
        if cfg_key not in conf:
            return conf

        new_conf = deepcopy(conf)
        for wrong_key, right_key in self._config_renamings.items():
            if wrong_key in new_conf[cfg_key]:
                new_conf[cfg_key][right_key] = new_conf[cfg_key].pop(wrong_key)
                if conflict == "warn":
                    warnings.warn(
                        "Altering config flag {} to {}".format(wrong_key, right_key)
                    )

        return new_conf

    def set_zero_config(self):
        """Set config such that radiative forcing and temperature output will be zero

        This method is intended as a convenience only, it does not handle everything in
        an obvious way. Adjusting the parameter settings still requires great care and
        may behave unepexctedly.
        """
        # zero_emissions is imported from scenarios module
        # TODO: setup MAGICC6 so it puts extra variables in right place and hence
        # warning about ignoring some data disappears
        zero_emissions.write(join(self.run_dir, self._scen_file_name), self.version)

        time = zero_emissions.filter(variable="Emissions|CH4", region="World")[
            "time"
        ].values
        no_timesteps = len(time)
        # value doesn't actually matter as calculations are done from difference but
        # chose sensible value nonetheless
        co2_conc_pi = 722
        co2_conc = co2_conc_pi * np.ones(no_timesteps)
        co2_conc_df = pd.DataFrame(
            {
                "time": time,
                "scenario": "idealised",
                "model": "unspecified",
                "climate_model": "unspecified",
                "variable": "Atmospheric Concentrations|CO2",
                "unit": "ppm",
                "todo": "SET",
                "region": "World",
                "value": co2_conc,
            }
        )
        co2_conc_writer = MAGICCData(co2_conc_df)
        co2_conc_filename = "HIST_CONSTANT_CO2_CONC.IN"
        co2_conc_writer.metadata = {
            "header": "Constant pre-industrial CO2 concentrations"
        }
        co2_conc_writer.write(join(self.run_dir, co2_conc_filename), self.version)

        ch4_conc_pi = 722
        ch4_conc = ch4_conc_pi * np.ones(no_timesteps)
        ch4_conc_df = pd.DataFrame(
            {
                "time": time,
                "scenario": "idealised",
                "model": "unspecified",
                "climate_model": "unspecified",
                "variable": "Atmospheric Concentrations|CH4",
                "unit": "ppb",
                "todo": "SET",
                "region": "World",
                "value": ch4_conc,
            }
        )
        ch4_conc_writer = MAGICCData(ch4_conc_df)
        ch4_conc_filename = "HIST_CONSTANT_CH4_CONC.IN"
        ch4_conc_writer.metadata = {
            "header": "Constant pre-industrial CH4 concentrations"
        }
        ch4_conc_writer.write(join(self.run_dir, ch4_conc_filename), self.version)

        fgas_conc_pi = 0
        fgas_conc = fgas_conc_pi * np.ones(no_timesteps)

        varname = "FGAS_CONC"
        fgas_conc_df = pd.DataFrame(
            {
                "time": time,
                "scenario": "idealised",
                "model": "unspecified",
                "climate_model": "unspecified",
                "variable": varname,
                "unit": "ppt",
                "todo": "SET",
                "region": "World",
                "value": fgas_conc,
            }
        )
        fgas_conc_writer = MAGICCData(fgas_conc_df)
        fgas_conc_filename = "HIST_ZERO_{}.IN".format(varname)
        fgas_conc_writer.metadata = {"header": "Zero concentrations"}
        fgas_conc_writer.write(join(self.run_dir, fgas_conc_filename), self.version)

        def_config = self.default_config
        tmp_nml = f90nml.Namelist({"nml_allcfgs": {"fgas_files_conc": 1}})
        fgas_files_conc_flag = list(
            self._fix_legacy_keys(tmp_nml, conflict="ignore")["nml_allcfgs"].keys()
        )[0]
        fgas_conc_files = [fgas_conc_filename] * len(
            def_config["nml_allcfgs"][fgas_files_conc_flag]
        )

        self.set_config(
            conflict="ignore",
            file_emisscen=self._scen_file_name,
            rf_initialization_method="ZEROSTARTSHIFT",
            rf_total_constantafteryr=10000,
            file_co2i_emis="",
            file_co2b_emis="",
            file_co2_conc=co2_conc_filename,
            co2_switchfromconc2emis_year=10000,
            file_ch4i_emis="",
            file_ch4b_emis="",
            file_ch4n_emis="",
            file_ch4_conc=ch4_conc_filename,
            ch4_switchfromconc2emis_year=10000,
            file_n2oi_emis="",
            file_n2ob_emis="",
            file_n2on_emis="",
            file_n2o_conc="",
            n2o_switchfromconc2emis_year=1750,
            file_noxi_emis="",
            file_noxb_emis="",
            file_noxi_ot="",
            file_noxb_ot="",
            file_noxt_rf="",
            file_soxnb_ot="",
            file_soxi_ot="",
            file_soxt_rf="",
            file_soxi_emis="",
            file_soxb_emis="",
            file_soxn_emis="",
            file_oci_emis="",
            file_ocb_emis="",
            file_oci_ot="",
            file_ocb_ot="",
            file_oci_rf="",
            file_ocb_rf="",
            file_bci_emis="",
            file_bcb_emis="",
            file_bci_ot="",
            file_bcb_ot="",
            file_bci_rf="",
            file_bcb_rf="",
            bcoc_switchfromrf2emis_year=1750,
            file_nh3i_emis="",
            file_nh3b_emis="",
            file_nmvoci_emis="",
            file_nmvocb_emis="",
            file_coi_emis="",
            file_cob_emis="",
            file_mineraldust_rf="",
            file_landuse_rf="",
            file_bcsnow_rf="",
            # rf_fgassum_scale=0,  # this appears to do nothing, hence the next two lines
            fgas_switchfromconc2emis_year=10000,
            rf_mhalosum_scale=0,
            stratoz_o3scale=0,
            rf_volcanic_scale=0,
            rf_solar_scale=0,
            mhalo_switchfromconc2emis_year=1750,
            fgas_files_conc=fgas_conc_files,
        )

    def _format_config(self, config_dict):
        # config_dict = self._fix_any_backwards_emissions_scen_key_in_config(config_dict)
        config_dict = self._convert_out_config_flags_to_integers(config_dict)

        return config_dict

    def _convert_out_config_flags_to_integers(self, config_dict):
        valid_out_flags = [
            "out_emissions",
            "out_gwpemissions",
            "out_sum_gwpemissions",
            "out_concentrations",
            "out_carboncycle",
            "out_forcing",
            "out_forcing_subannual",
            "out_temperature",
            "out_temperature_subannual",
            "out_sealevel",
            "out_parameters",
            "out_misc",
            "out_lifetimes",
            "out_timeseriesmix",
            "out_rcpdata",
            "out_summaryidx",
            "out_tempoceanlayers",
            "out_oceanarea",
            "out_heatuptake",
            "out_warnings",
            "out_precipinput",
            "out_aogcmtuning",
            "out_ccycletuning",
            "out_observationaltuning",
            "out_keydata_1",
            "out_keydata_2",
            "out_inverseemis",
            "out_surfaceforcing",
            "out_permafrost",
            "out_allowanydynamicvars",
        ]
        for key in valid_out_flags:
            if key in config_dict:
                # MAGICC expects 1 and 0 instead of True/False
                config_dict[key] = 1 if config_dict[key] else 0

        return config_dict

    def set_years(self, startyear=1765, endyear=2100):
        """
        Set the start and end dates of the simulations.

        Parameters
        ----------
        startyear : int
            Start year of the simulation

        endyear : int
            End year of the simulation

        Returns
        -------
        dict
            The contents of the namelist
        """
        # TODO: test altering stepsperyear, I think 1, 2 and 24 should all work
        return self.set_config(
            "MAGCFG_NMLYEARS.CFG",
            "nml_years",
            endyear=endyear,
            startyear=startyear,
            stepsperyear=12,
        )

    def set_output_variables(self, write_ascii=True, write_binary=False, **kwargs):
        """Set the output configuration, minimising output as much as possible

        There are a number of configuration parameters which control which variables
        are written to file and in which format. Limiting the variables that are
        written to file can greatly speed up the running of MAGICC. By default,
        calling this function without specifying any variables will disable all output
        by setting all of MAGICC's ``out_xx`` flags to ``0``.

        This convenience function should not be confused with ``set_config`` or
        ``update_config`` which allow the user to set/update the configuration flags
        directly, without the more convenient syntax and default behaviour provided by
        this function.

        Parameters
        ----------
        write_ascii : bool
            If true, MAGICC is configured to write output files as human readable ascii files.

        write_binary : bool
            If true, MAGICC is configured to write binary output files. These files are much faster
            to process and write, but are not human readable.

        **kwargs:
            List of variables to write out. A list of possible options are as follows. This
            may not be a complete list.

            'emissions',
            'gwpemissions',
            'sum_gwpemissions',
            'concentrations',
            'carboncycle',
            'forcing',
            'surfaceforcing',
            'permafrost',
            'temperature',
            'sealevel',
            'parameters',
            'misc',
            'lifetimes',
            'timeseriesmix',
            'rcpdata',
            'summaryidx',
            'inverseemis',
            'tempoceanlayers',
            'oceanarea',
            'heatuptake',
            'warnings',
            'precipinput',
            'aogcmtuning',
            'ccycletuning',
            'observationaltuning',
            'keydata_1',
            'keydata_2'
        """

        assert (
            write_ascii or write_binary
        ), "write_binary and/or write_ascii must be configured"
        if write_binary and write_ascii:
            ascii_binary = "BOTH"
        elif write_ascii:
            ascii_binary = "ASCII"
        else:
            ascii_binary = "BINARY"

        # defaults
        outconfig = {
            "out_emissions": 0,
            "out_gwpemissions": 0,
            "out_sum_gwpemissions": 0,
            "out_concentrations": 0,
            "out_carboncycle": 0,
            "out_forcing": 0,
            "out_surfaceforcing": 0,
            "out_permafrost": 0,
            "out_temperature": 0,
            "out_sealevel": 0,
            "out_parameters": 0,
            "out_misc": 0,
            "out_timeseriesmix": 0,
            "out_rcpdata": 0,
            "out_summaryidx": 0,
            "out_inverseemis": 0,
            "out_tempoceanlayers": 0,
            "out_heatuptake": 0,
            "out_ascii_binary": ascii_binary,
            "out_warnings": 0,
            "out_precipinput": 0,
            "out_aogcmtuning": 0,
            "out_ccycletuning": 0,
            "out_observationaltuning": 0,
            "out_keydata_1": 0,
            "out_keydata_2": 0,
        }
        if self.version == 7:
            outconfig["out_oceanarea"] = 0
            outconfig["out_lifetimes"] = 0

        for kw in kwargs:
            val = 1 if kwargs[kw] else 0  # convert values to 0/1 instead of booleans
            outconfig["out_" + kw.lower()] = val

        self.update_config(**outconfig)

    def get_executable(self):
        return config["executable_{}".format(self.version)]

    def diagnose_tcr_ecs(self, **kwargs):
        """Diagnose TCR and ECS

        The transient climate response (TCR), is the global-mean temperature response
        at time at which atmopsheric |CO2| concentrations double in a scenario where
        atmospheric |CO2| concentrations are increased at 1% per year from
        pre-industrial levels.

        The equilibrium climate sensitivity (ECS), is the equilibrium global-mean
        temperature response to an instantaneous doubling of atmospheric |CO2|
        concentrations.

        As MAGICC has no hysteresis in its equilibrium response to radiative forcing,
        we can diagnose TCR and ECS with one experiment. However, please note that
        sometimes the run length won't be long enough to allow MAGICC's oceans to
        fully equilibrate and hence the ECS value might not be what you expect (it
        should match the value of ``core_climatesensitivity``).

        Parameters
        ----------
        **kwargs
            parameter values to use in the diagnosis e.g. ``core_climatesensitivity=4``

        Returns
        -------
        dict
            Dictionary with keys: "ecs" - the diagnosed ECS; "tcr" - the diagnosed
            TCR; "timeseries" - the relevant model input and output timeseries used in
            the experiment i.e. atmospheric |CO2| concentrations, total radiative
            forcing and global-mean surface temperature
        """
        self._diagnose_tcr_ecs_config_setup(**kwargs)
        timeseries = self.run(
            scenario=None,
            only=[
                "Atmospheric Concentrations|CO2",
                "Radiative Forcing",
                "Surface Temperature",
            ],
        )
        tcr, ecs = self._get_tcr_ecs_from_diagnosis_results(timeseries)
        return {"tcr": tcr, "ecs": ecs, "timeseries": timeseries}

    def _diagnose_tcr_ecs_config_setup(self, **kwargs):
        self.set_years(
            startyear=1750, endyear=4200
        )  # 4200 seems to be the max I can push too without an error

        self.update_config(
            FILE_CO2_CONC="TCRECS_CO2_CONC.IN",
            CO2_SWITCHFROMCONC2EMIS_YEAR=30000,
            RF_TOTAL_RUNMODUS="CO2",
            RF_TOTAL_CONSTANTAFTERYR=2000,
            **kwargs,
        )

    def _get_tcr_ecs_from_diagnosis_results(self, results_tcr_ecs_run):
        global_co2_concs = results_tcr_ecs_run.filter(
            variable="Atmospheric Concentrations|CO2", region="World"
        )
        tcr_time, ecs_time = self._get_tcr_ecs_yr_from_CO2_concs(global_co2_concs)

        global_total_rf = results_tcr_ecs_run.filter(
            variable="Radiative Forcing", region="World"
        )
        self._check_tcr_ecs_total_RF(
            global_total_rf, tcr_time=tcr_time, ecs_time=ecs_time
        )

        global_temp = results_tcr_ecs_run.filter(
            variable="Surface Temperature", region="World"
        )
        self._check_tcr_ecs_temp(global_temp)

        tcr = float(global_temp.filter(time=tcr_time).values.squeeze())
        ecs = float(global_temp.filter(time=ecs_time).values.squeeze())

        return tcr, ecs

    def _get_tcr_ecs_yr_from_CO2_concs(self, df_co2_concs):
        co2_concs = df_co2_concs.timeseries()
        co2_conc_0 = co2_concs.iloc[0, 0]
        t_start = co2_concs.columns.min()
        t_end = co2_concs.columns.max()

        t_start_rise = co2_concs.iloc[
            :, co2_concs.values.squeeze() > co2_conc_0
        ].columns[0] - relativedelta(years=1)
        tcr_time = t_start_rise + relativedelta(years=70)

        spin_up_co2_concs = (
            _filter_time_range(df_co2_concs, lambda x: t_start <= x <= t_start_rise)
            .timeseries()
            .values.squeeze()
        )
        if not (spin_up_co2_concs == co2_conc_0).all():
            raise ValueError(
                "The TCR/ECS CO2 concs look wrong, they are not constant before they start rising"
            )

        actual_rise_co2_concs = (
            _filter_time_range(df_co2_concs, lambda x: t_start_rise <= x <= tcr_time)
            .timeseries()
            .values.squeeze()
        )
        # this will blow up if we switch to diagnose tcr/ecs with a monthly run...
        expected_rise_co2_concs = co2_conc_0 * 1.01 ** np.arange(71)
        rise_co2_concs_correct = np.isclose(
            actual_rise_co2_concs, expected_rise_co2_concs
        ).all()
        if not rise_co2_concs_correct:
            raise ValueError("The TCR/ECS CO2 concs look wrong during the rise period")

        co2_conc_final = max(expected_rise_co2_concs)
        eqm_co2_concs = (
            _filter_time_range(df_co2_concs, lambda x: tcr_time <= x <= t_end)
            .timeseries()
            .values.squeeze()
        )
        if not np.isclose(eqm_co2_concs, co2_conc_final).all():
            raise ValueError(
                "The TCR/ECS CO2 concs look wrong, they are not constant after 70 years of rising"
            )

        ecs_time = df_co2_concs["time"].iloc[-1]

        return tcr_time, ecs_time

    def _check_tcr_ecs_total_RF(self, df_total_rf, tcr_time, ecs_time):
        total_rf = df_total_rf.timeseries()
        total_rf_max = total_rf.values.squeeze().max()

        t_start = total_rf.columns.min()
        t_end = total_rf.columns.max()
        t_start_rise = tcr_time - relativedelta(years=70)

        spin_up_rf = (
            _filter_time_range(df_total_rf, lambda x: t_start <= x <= t_start_rise)
            .timeseries()
            .values.squeeze()
        )
        if not (spin_up_rf == 0).all():
            raise ValueError(
                "The TCR/ECS total radiative forcing looks wrong, it is not all zero before concentrations start rising"
            )

        eqm_rf = (
            _filter_time_range(df_total_rf, lambda x: tcr_time <= x <= t_end)
            .timeseries()
            .values.squeeze()
        )
        if not (eqm_rf == total_rf_max).all():
            raise ValueError(
                "The TCR/ECS total radiative forcing looks wrong, it is not constant after concentrations are constant"
            )

    def _check_tcr_ecs_temp(self, df_temp):
        tmp_vls = df_temp.timeseries().values.squeeze()
        tmp_minus_previous_yr = tmp_vls[1:] - tmp_vls[:-1]
        if not np.all(tmp_minus_previous_yr >= 0):
            raise ValueError(
                "The TCR/ECS surface temperature looks wrong, it decreases"
            )

    def set_emission_scenario_setup(self, scenario, config_dict):
        """Set the emissions flags correctly.

        Parameters
        ----------
        scenario : :obj:`pymagicc.io.MAGICCData`
            Scenario to run.

        config_dict : dict
            Dictionary with current input configurations which is to be validated and
            updated where necessary.

        Returns
        -------
        dict
            Updated configuration
        """
        self.write(scenario, self._scen_file_name)
        emis_flag = list(
            self._fix_legacy_keys(
                f90nml.Namelist({"nml_allcfgs": {"file_emisscen": "junk"}}),
                conflict="ignore",
            )["nml_allcfgs"].keys()
        )[0]
        config_dict[emis_flag] = self._scen_file_name

        return config_dict

    def _check_config(self):
        """
        Check config above and beyond those checked by ``self.check_config``
        """
        pass


class MAGICC6(MAGICCBase):
    version = 6
    _scen_file_name = "SCENARIO.SCEN"

    _config_renamings = {
        "file_emisscen": "file_emissionscenario",
        "fgas_files_conc": "file_fgas_conc",
        "mhalo_switchfromconc2emis_year": "mhalo_switch_conc2emis_yr",
    }

    @property
    def default_config(self):
        """
        Default configuration to use in a run
        """
        base = f90nml.read(join(self.run_dir, "MAGCFG_DEFAULTALL_69.CFG"))
        user = f90nml.read(join(self.run_dir, "MAGCFG_USER.CFG"))
        self._default_config = deepcopy(base)
        self._default_config.update(user)

        return self._default_config

    def _check_tcr_ecs_total_RF(self, df_total_rf, tcr_time, ecs_time):
        super()._check_tcr_ecs_total_RF(df_total_rf, tcr_time, ecs_time)
        # can be more careful with checks MAGICC6 only has logarithmic CO2 forcing
        # i.e. linear rise in forcing
        total_rf = df_total_rf.timeseries()
        total_rf_max = total_rf.values.squeeze().max()

        t_start_rise = tcr_time - relativedelta(years=70)

        actual_rise_rf = (
            _filter_time_range(df_total_rf, lambda x: t_start_rise <= x <= tcr_time)
            .timeseries()
            .values.squeeze()
        )

        # this will blow up if we switch to diagnose tcr/ecs with a monthly run...
        expected_rise_rf = total_rf_max / 70.0 * np.arange(71)
        rise_rf_correct = np.isclose(actual_rise_rf, expected_rise_rf).all()
        if not rise_rf_correct:
            raise ValueError(
                "The TCR/ECS total radiative forcing looks wrong during the rise period"
            )

    def _check_config(self):
        cfg = self.update_config()
        if "file_emissionscenario" in cfg["nml_allcfgs"]:
            if cfg["nml_allcfgs"]["file_emissionscenario"].endswith("SCEN7"):
                self._check_failed("MAGICC6 cannot run SCEN7 files")


class MAGICC7(MAGICCBase):
    version = 7

    _config_renamings = {
        "file_emissionscenario": "file_emisscen",
        "file_fgas_conc": "fgas_files_conc",
        "mhalo_switch_conc2emis_yr": "mhalo_switchfromconc2emis_year",
    }

    def create_copy(self):
        """
        Initialises a temporary directory structure and copy of MAGICC
        configuration files and binary.

        This will also overwrite the value of all ``file_tuningmodel_x`` flags to
        ensure that Pymagicc's configurations will be read. If ``self.strict``, this
        will also overwrite the value of all ``file_emisscen_x`` flags to ensure that
        only Pymagicc's scenario input is used. This overwrite behaviour can be
        removed once the MAGICC7 binary is publicly released as we can then create a
        Pymagicc specific MAGCFG_USER.CFG rather than relying on whatever is in the
        user's current copy.
        """
        super(MAGICC7, self).create_copy()
        self.update_config(
            "MAGCFG_USER.CFG",
            **{
                "file_tuningmodel_1": "PYMAGICC",
                "file_tuningmodel_2": "USER",
                "file_tuningmodel_3": "USER",
                "file_tuningmodel_4": "USER",
                "file_tuningmodel_5": "USER",
                "file_tuningmodel_6": "USER",
                "file_tuningmodel_7": "USER",
                "file_tuningmodel_8": "USER",
                "file_tuningmodel_9": "USER",
                "file_tuningmodel_10": "USER",
            },
        )
        if self.strict:
            self.update_config(
                "MAGCFG_USER.CFG",
                **{
                    "file_emisscen_2": "NONE",
                    "file_emisscen_3": "NONE",
                    "file_emisscen_4": "NONE",
                    "file_emisscen_5": "NONE",
                    "file_emisscen_6": "NONE",
                    "file_emisscen_7": "NONE",
                    "file_emisscen_8": "NONE",
                },
            )

    def _diagnose_tcr_ecs_config_setup(self, **kwargs):
        super()._diagnose_tcr_ecs_config_setup(**kwargs)
        # also need to lock CH4 and N2O in case OLBL forcing mode is being used
        self.update_config(
            FILE_CH4_CONC="TCRECS_CH4_CONC.IN",
            CH4_SWITCHFROMCONC2EMIS_YEAR=30000,
            FILE_N2O_CONC="TCRECS_N2O_CONC.IN",
            N2O_SWITCHFROMCONC2EMIS_YEAR=30000,
        )

    def _check_config(self):
        pass


def _filter_time_range(scmdf, filter_func):
    # TODO: move into openscm
    tdf = scmdf.timeseries()
    tdf = tdf.iloc[:, tdf.columns.map(filter_func)]
    return MAGICCData(tdf)



# Copyright (c) 2017 pymagicc authors:
#   Robert Gieseke <robert.gieseke@pik-potsdam.de>
# Free software under GNU Affero General Public License v3, see LICENSE
#
# The compiled MAGICC binary (http://www.magicc.org/download6) by Tom Wigley,
# Sarah Raper, and Malte Meinshausen included in this package is licensed under
# a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported
# License (https://creativecommons.org/licenses/by-nc-sa/3.0/)
from ._version import get_versions
from .core import MAGICC6, MAGICC7  # noqa
from .io import MAGICCData, read_scen_file  # noqa
from .scenarios import rcp26, rcp45, rcp60, rcp85, rcps, zero_emissions  # noqa


__version__ = get_versions()["version"]
del get_versions


def run(scenario, magicc_version=6, **kwargs):
    """
    Run a MAGICC scenario and return output data and (optionally) config parameters.

    As a reminder, putting ``out_parameters=1`` will cause MAGICC to write out its
    parameters into ``out/PARAMETERS.OUT`` and they will then be read into
    ``output.metadata["parameters"]`` where ``output`` is the returned object.

    Parameters
    ----------
    scenario : :obj:`pymagicc.io.MAGICCData`
        Scenario to run

    magicc_version : int
        MAGICC version to use for the run

    **kwargs
        Parameters overwriting default parameters

    Raises
    ------
    ValueError
        If the magicc_version is not available

    Returns
    -------
    output : :obj:`pymagicc.io.MAGICCData`
        Output of the run with the data in the ``df`` attribute and parameters and
        other metadata in the ``metadata attribute``
    """
    if magicc_version == 6:
        magicc_cls = MAGICC6
    elif magicc_version == 7:
        magicc_cls = MAGICC7
    else:
        raise ValueError("MAGICC version {} is not available".format(magicc_version))

    with magicc_cls() as magicc:
        results = magicc.run(scenario=scenario, **kwargs)

    return results
